@@ -1,6 +1,6 @@
 # Orca
 #
-# Copyright 2005 Sun Microsystems Inc.
+# Copyright 2005-2007 Sun Microsystems Inc.
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Library General Public
@@ -24,14 +24,22 @@ machine to speak.  Each speech server provides a set of known
 voices (identified by name) which can be combined with various
 attributes to create aural style sheets."""
 
-__id__        = "$Id: speechserver.py 1036 2006-06-10 00:54:48Z wwalker $"
-__version__   = "$Revision: 1036 $"
-__date__      = "$Date: 2006-06-09 20:54:48 -0400 (Fri, 09 Jun 2006) $"
-__copyright__ = "Copyright (c) 2005-2006 Sun Microsystems Inc."
+__id__        = "$Id: speechserver.py 2528 2007-07-09 14:41:22Z wwalker $"
+__version__   = "$Revision: 2528 $"
+__date__      = "$Date: 2007-07-09 10:41:22 -0400 (Mon, 09 Jul 2007) $"
+__copyright__ = "Copyright (c) 2005-2007 Sun Microsystems Inc."
 __license__   = "LGPL"
 
+import logging
+import keynames
+import settings
+import orca
+log = logging.getLogger("speech")
+
 import debug
 
+from orca_i18n import _           # for gettext support
+
 class VoiceFamily(dict):
     """Holds the family description for a voice."""
 
@@ -81,7 +89,8 @@ class SayAllContext:
         self.currentOffset = startOffset
         self.endOffset     = endOffset
 
-class SpeechServer:
+
+class SpeechServer(object):
 
     """Provides speech server abstraction."""
 
@@ -170,6 +179,36 @@ class SpeechServer:
         """
         pass
 
+    def speakKeyEvent(self, event_string, type):
+        """Speaks a key event immediately.
+
+        Arguments:
+        - event_string: string representing the key event as defined by
+                        input_event.KeyboardEvent.
+        - type:         key event type as one of orca.KeyEventType constants.
+
+        """
+        if type == orca.KeyEventType.PRINTABLE and \
+               event_string.decode("UTF-8").isupper():
+            voice = settings.voices[settings.UPPERCASE_VOICE]
+        else:
+            voice = settings.voices[settings.DEFAULT_VOICE]
+
+        # Check to see if there are localized words to be spoken for
+        # this key event.
+        #
+        event_string = keynames.getKeyName(event_string)
+
+        if type == orca.KeyEventType.LOCKING_LOCKED:
+            event_string += " " + _("on")
+        elif type == orca.KeyEventType.LOCKING_UNLOCKED:
+            event_string += " " + _("off")
+
+        debug.println(debug.LEVEL_INFO, "SPEECH OUTPUT: '" + event_string +"'")
+        log.info("'%s'" % event_string)
+
+        self.speak(event_string, acss=voice)
+
     def speakUtterances(self, list, acss=None, interrupt=True):
         """Speaks the given list of utterances immediately.
 
@@ -201,8 +240,8 @@ class SpeechServer:
         pass
 
     def isSpeaking(self):
-	""""Returns True if the system is currently speaking."""
-	return False
+        """"Returns True if the system is currently speaking."""
+        return False
 
     def sayAll(self, utteranceIterator, progressCallback):
         """Iterates through the given utteranceIterator, speaking
@@ -221,6 +260,7 @@ class SpeechServer:
         for [context, acss] in utteranceIterator:
             debug.println(debug.LEVEL_INFO,
                           "SPEECH OUTPUT: '" + context.utterance + "'")
+            log.info("'%s'" % context.utterance)
             self.speak(context.utterance, acss)
 
     def increaseSpeechRate(self, step=5):
